(*
License: LGPL (see file COPYING.LESSER)
*)
theory IsaFoR_Examples
imports WDP_Transformation
begin

text \<open>
  Equality modulo ``constructor symbols'' (more specifically, \emph{all} symbols that are not in
  the set of defined symbols.
\<close>
locale eqmodc_defs =
  fixes \<D> :: "'f sig" -- \<open>defined symbols\<close>
begin

abbreviation "ccap \<equiv> cap_till_funas \<D> :: ('f, 'v) term \<Rightarrow> _"
abbreviation "dmax \<equiv> uncap_till_funas \<D> :: ('f, 'v) term \<Rightarrow> _"
abbreviation "compose_ccap \<equiv> compose_cap_till_funas \<D> :: ('f, 'v) term \<Rightarrow> _"

lemma fill_holes_dmax_eq:
  "ground_mctxt C \<Longrightarrow> num_holes C = length (dmax t) \<Longrightarrow> funas_mctxt C \<subseteq> - \<D> \<Longrightarrow>
    s = fill_holes C (dmax t) \<Longrightarrow> dmax s = dmax t"
  by auto

text \<open>Equality modulo constructor contexts\<close>
definition "eqmodc s t \<longleftrightarrow> (\<exists>C. ground_mctxt C \<and> funas_mctxt C \<subseteq> - \<D> \<and> t =\<^sub>f (C, dmax s))"

notation eqmodc (infix "=\<^sub>C" 50)

lemma eqmodc_refl:
  "t =\<^sub>C t"
  using split_term_eqf [of t "if_Fun_in_set \<D>"] and cap_till_funas [of "\<D>" t]
  by (auto simp: eqmodc_def)

lemma eqmodc_sym:
  assumes "s =\<^sub>C t"
  shows "t =\<^sub>C s"
proof -
  obtain C where "ground_mctxt C" and "funas_mctxt C \<subseteq> - \<D>" and "t =\<^sub>f (C, dmax s)"
    using assms by (auto simp: eqmodc_def)
  with fill_holes_dmax_eq have [simp]: "dmax t = dmax s" by (metis eqfE)
  have "s =\<^sub>f (ccap s, dmax s)" by (rule split_term_eqf)
  then have "s =\<^sub>f (ccap s, dmax t)" by simp
  moreover have "funas_mctxt (ccap s) \<subseteq> - \<D>" by (metis Compl_iff cap_till_funas subsetI)
  ultimately show ?thesis by (auto simp: eqmodc_def)
qed

lemma eqmodc_trans:
  "s =\<^sub>C t \<Longrightarrow> t =\<^sub>C u \<Longrightarrow> s =\<^sub>C u"
  by (auto simp: eqmodc_def) (metis eqfE fill_holes_dmax_eq)

lemma qrstep_eqmodc:
  assumes defs_R: "{fn. defined R fn} \<subseteq> \<D>"
    and wf_R: "wf_trs R"
  assumes "(s, t) \<in> qrstep nfs Q R" and "s =\<^sub>C u"
  shows "\<exists>v. t =\<^sub>C v \<and> (u, v) \<in> qrstep nfs Q R"
proof -
  from assms obtain C l r \<sigma>
    where args: "\<forall>u \<lhd> l \<cdot> \<sigma>. u \<in> NF_terms Q" and nfs: "NF_subst nfs (l, r) \<sigma> Q"
    and rule: "(l, r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" by auto
  then obtain f and n where defined: "(f, n) \<in> \<D>"
    and root: "root (l \<cdot> \<sigma>) = Some (f, n)" using wf_R and defs_R
    by (cases l, auto simp: wf_trs_def defined_def)
       (metis contra_subsetD defined_def defs_R mem_Collect_eq root.simps(2))
  from in_uncap_till_funas [OF root, of "\<D>", OF defined s] obtain i and D
    where i: "i < length (dmax s)" "dmax s ! i = D\<langle>l \<cdot> \<sigma>\<rangle>"
      and C: "mctxt_of_ctxt C = compose_ccap s i (mctxt_of_ctxt D)" by blast
  def ss\<^sub>1 \<equiv> "take i (dmax s)"
  def ss\<^sub>2 \<equiv> "drop (Suc i) (dmax s)"
  have [simp]: "length ss\<^sub>1 = i" using i by (simp add: ss\<^sub>1_def)
  have dmax_s: "dmax s = ss\<^sub>1 @ D\<langle>l \<cdot> \<sigma>\<rangle> # ss\<^sub>2"
    using i by (auto simp: ss\<^sub>1_def ss\<^sub>2_def dest: id_take_nth_drop)
  have "t =\<^sub>f (compose_ccap s i (mctxt_of_ctxt D), [r \<cdot> \<sigma>])"
    using t unfolding C [symmetric] by (metis mctxt_of_ctxt)
  from eqfE [OF this]
    have "t =\<^sub>f (ccap s, ss\<^sub>1 @ D\<langle>r \<cdot> \<sigma>\<rangle> # ss\<^sub>2)"
    using fill_holes_compose_cap_till [of i "if_Fun_in_set \<D>" s "mctxt_of_ctxt D" "[r \<cdot> \<sigma>]"] and i
    by (auto simp: split_term_num_holes ss\<^sub>1_def ss\<^sub>2_def)
  from eqfE [OF this]
    have t: "t = fill_holes (ccap s) (ss\<^sub>1 @ D\<langle>r \<cdot> \<sigma>\<rangle> # ss\<^sub>2)"
    and *: "num_holes (ccap s) = length (ss\<^sub>1 @ D\<langle>r \<cdot> \<sigma>\<rangle> # ss\<^sub>2)" by simp_all
  from \<open>s =\<^sub>C u\<close> obtain E
    where "u =\<^sub>f (E, dmax s)" and E: "ground_mctxt E" "funas_mctxt E \<subseteq> - \<D>" by (auto simp: eqmodc_def)
  then have u: "u = fill_holes E (dmax s)"
    and "num_holes E = length (dmax s)" by (auto dest: eqfE)
  from uncap_till_funas_fill_holes_cancel [OF this(2), of "\<D>"] and E
    have dmax_u: "dmax u = dmax s" by (simp add: u)

  have "dmax t = concat (map (dmax) (ss\<^sub>1 @ D\<langle>r \<cdot> \<sigma>\<rangle> # ss\<^sub>2))"
    unfolding t using * by simp
  also have "\<dots> = ss\<^sub>1 @ dmax (D\<langle>r \<cdot> \<sigma>\<rangle>) @ ss\<^sub>2"
    by (simp add: ss\<^sub>1_def ss\<^sub>2_def drop_map [symmetric] take_map [symmetric])
       (simp add: drop_map take_map)
  finally have dmax_t: "dmax t = ss\<^sub>1 @ dmax (D\<langle>r \<cdot> \<sigma>\<rangle>) @ ss\<^sub>2" .

  let ?D = "ccap (D\<langle>r \<cdot> \<sigma>\<rangle>)"
  let ?u = "dmax (D\<langle>r \<cdot> \<sigma>\<rangle>)"
  def G \<equiv> "compose_mctxt (ccap u) i ?D"
  def v \<equiv> "fill_holes G (ss\<^sub>1 @ ?u @ ss\<^sub>2)"

  have **: "fill_holes (ccap u) (ss\<^sub>1 @ D\<langle>r \<cdot> \<sigma>\<rangle> # ss\<^sub>2) =\<^sub>f (G, ss\<^sub>1 @ ?u @ ss\<^sub>2)"
  proof (unfold G_def, rule compose_mctxt_sound [OF _ split_term_eqf])
    show "fill_holes (ccap u) (ss\<^sub>1 @ D\<langle>r \<cdot> \<sigma>\<rangle> # ss\<^sub>2) =\<^sub>f (ccap u, ss\<^sub>1 @ D\<langle>r \<cdot> \<sigma>\<rangle> # ss\<^sub>2)"
      by (rule eq_fill.intros) (simp_all add: split_term_num_holes dmax_u dmax_s)
  next
    show "i = length ss\<^sub>1" using i by (simp add: ss\<^sub>1_def)
  qed
  then have "v =\<^sub>f (G, dmax t)" by (auto simp: dmax_t v_def dest: eqfE)
  moreover have "ground_mctxt G" by (auto simp: G_def)
  moreover
  have "funas_mctxt G \<subseteq> - \<D>"
    unfolding G_def
    unfolding funas_mctxt_compose_mctxt [of i "ccap u" ?D, unfolded split_term_num_holes dmax_u, OF i(1)]
    using cap_till_funas [of "\<D>"] by auto
  moreover
  have "(u, v) \<in> qrstep nfs Q R"
  proof -
    from fill_holes_ctxt_main [of "ccap u" ss\<^sub>1 ss\<^sub>2] obtain C
      where [simp]: "\<And>t. fill_holes (ccap u) (ss\<^sub>1 @ t # ss\<^sub>2) = C\<langle>t\<rangle>"
      using * by (auto simp: split_term_num_holes dmax_u)
    have "u = fill_holes (ccap u) (ss\<^sub>1 @ D\<langle>l \<cdot> \<sigma>\<rangle> # ss\<^sub>2)"
      by (metis dmax_s dmax_u fill_holes_cap_till_uncap_till_id)
    then have "u = (C \<circ>\<^sub>c D)\<langle>l \<cdot> \<sigma>\<rangle>" by simp
    have "v = fill_holes (ccap u) (ss\<^sub>1 @ D\<langle>r \<cdot> \<sigma>\<rangle> # ss\<^sub>2)"
      using ** by (auto simp: v_def dest: eqfE)
    then have "v = (C \<circ>\<^sub>c D)\<langle>r \<cdot> \<sigma>\<rangle>" by simp
    show ?thesis by (rule qrstepI) fact+
  qed
  ultimately
  show ?thesis unfolding eqmodc_def by blast
qed

end

end

