theory Lambda_Z_Nominal
imports
  "$AFP/Nominal2/Nominal2"
  "$AFP/Abstract-Rewriting/Abstract_Rewriting"
begin

atom_decl name

nominal_datatype "term" =
  Var name
| App "term" "term"
| Abs x::name t::"term" binds x in t

nominal_function subst
where
  "subst x s (Var y) = (if x = y then s else Var y)"
| "subst x s (App t u) = App (subst x s t) (subst x s u)"
| "atom y \<sharp> (x, s) \<Longrightarrow> subst x s (Abs y t) = Abs y (subst x s t)"
apply (simp add: eqvt_def subst_graph_aux_def)
apply (rule TrueI)
  using [[simproc del: alpha_lst]]
  apply (auto)
  apply (rule_tac y="b" and c="(a, aa)" in term.strong_exhaust)
  apply (blast)+
  apply (simp_all add: fresh_star_def fresh_Pair_elim)
  apply (erule_tac c="(xa, sa)" in Abs_lst1_fcb2)
  apply (simp_all add: Abs_fresh_iff)
  apply (simp add: fresh_star_def fresh_Pair)
  apply (simp add: eqvt_at_def)
  apply (simp add: perm_supp_eq fresh_star_Pair)
  apply (simp add: eqvt_at_def)
  apply (simp add: perm_supp_eq fresh_star_Pair)
done
nominal_termination (eqvt) by lexicographic_order

lemma supp_fv_term_conv: "supp t = fv_term t"
by (nominal_induct t rule: term.strong_induct) (auto simp: term.fv_defs term.supp)

lemma fresh_subst_id [simp]:
  "atom x \<sharp> t \<Longrightarrow> subst x s t = t"
by (nominal_induct t avoiding: x s rule: term.strong_induct) (auto simp: fresh_at_base)

lemma fresh_fact:
  "atom z \<sharp> s \<Longrightarrow> z = y \<or> atom z \<sharp> t \<Longrightarrow> atom z \<sharp> subst y s t"
by (nominal_induct t avoiding: z y s rule: term.strong_induct) auto

text \<open>The substitution lemma.\<close>
lemma subst_subst:  
  assumes "x \<noteq> y" and "atom x \<sharp> u"
  shows "subst y u (subst x s t) = subst x (subst y u s) (subst y u t)"
using assms by (nominal_induct t avoiding: x y u s rule: term.strong_induct) (auto simp: fresh_fact)

inductive_set Beta
where
  "(App (Abs x s) t, subst x t s) \<in> Beta"
| "(s, t) \<in> Beta \<Longrightarrow> (App s u, App t u) \<in> Beta"
| "(s, t) \<in> Beta \<Longrightarrow> (App u s, App u t) \<in> Beta"
| "(s, t) \<in> Beta \<Longrightarrow> (Abs x s, Abs x t) \<in> Beta"

equivariance Betap
lemmas Beta_eqvt = Betap.eqvt [to_set]

(*
nominal_inductive Betap .
print_theorems
lemmas Beta_strong_induct = Betap.strong_induct [to_set]
thm Beta_strong_induct Beta.induct
*)

abbreviation beta (infix "\<rightarrow>\<^sub>\<beta>" 50)
where
  "s \<rightarrow>\<^sub>\<beta> t \<equiv> (s, t) \<in> Beta"

abbreviation betas (infix "\<rightarrow>\<^sub>\<beta>\<^sup>*" 50)
where
  "s \<rightarrow>\<^sub>\<beta>\<^sup>* t \<equiv> (s, t) \<in> Beta\<^sup>*"

nominal_function abs_cases :: "term \<Rightarrow> term \<Rightarrow> term \<Rightarrow> term"
where
  "atom x \<sharp> u \<Longrightarrow> abs_cases (Abs x s') u v = subst x u s'"
| "abs_cases (Var x) u v = v"
| "abs_cases (App s t) u v = v"
  unfolding eqvt_def abs_cases_graph_aux_def
  apply(simp)
  apply(rule TrueI)
  using [[simproc del: alpha_lst]]
  apply (auto)
  apply (rule_tac y="a" and c="aa" in term.strong_exhaust)
  apply (blast)+
  apply (simp add: fresh_star_def fresh_Pair_elim)
apply (erule_tac c="ua" in Abs_lst1_fcb2')
apply (simp add: fresh_fact)
apply (simp add: fresh_star_def fresh_Pair fresh_Unit)
apply (auto simp add: subst.eqvt perm_supp_eq)
done
nominal_termination (eqvt) by lexicographic_order

nominal_function lambda_bullet :: "term \<Rightarrow> term" ("_\<^sup>\<bullet>" [1000] 999)
where
  "(Var x)\<^sup>\<bullet> = Var x"
| "(Abs x t)\<^sup>\<bullet> = Abs x (t\<^sup>\<bullet>)"
| "(App s t)\<^sup>\<bullet> = abs_cases (s\<^sup>\<bullet>) (t\<^sup>\<bullet>) (App (s\<^sup>\<bullet>) (t\<^sup>\<bullet>))"
apply (simp add: eqvt_def lambda_bullet_graph_aux_def)
  apply(rule TrueI)
  using [[simproc del: alpha_lst]]
  apply (auto)
  apply (rule_tac y="x" and c="()" in term.strong_exhaust)
  apply (blast)+
apply (erule_tac c="()" in Abs_lst1_fcb2')
apply (auto simp add: Abs_fresh_iff)
apply (simp add: fresh_star_def fresh_Unit)
apply (simp add: eqvt_at_def)
apply (simp add: eqvt_at_def)
done
nominal_termination (eqvt) by lexicographic_order

lemma App_betas:
  assumes "s \<rightarrow>\<^sub>\<beta>\<^sup>* t" and "u \<rightarrow>\<^sub>\<beta>\<^sup>* v"
  shows "App s u \<rightarrow>\<^sub>\<beta>\<^sup>* App t v"
using assms(1)
proof (induct)
  case base
  show ?case using assms(2) by (induct) (auto intro: Beta.intros rtrancl_into_rtrancl)
next
  case (step a b)
  then show ?case by (auto intro: Beta.intros rtrancl_into_rtrancl)
qed

lemma Abs_betas:
  assumes "s \<rightarrow>\<^sub>\<beta>\<^sup>* t"
  shows "Abs x s \<rightarrow>\<^sub>\<beta>\<^sup>* Abs x t"
using assms by (induct) (auto intro: Beta.intros rtrancl_into_rtrancl)

lemma self:
  "t \<rightarrow>\<^sub>\<beta>\<^sup>* t\<^sup>\<bullet>"
proof (nominal_induct t rule: term.strong_induct)
  case (App t u)
  then show ?case
    by (cases "t\<^sup>\<bullet>" rule: term.strong_exhaust [of _ _ "u\<^sup>\<bullet>"])
       (auto intro: App_betas Beta.intros rtrancl_into_rtrancl simp: fresh_star_def)
qed (auto intro: Abs_betas)

lemma fresh_atom_bullet:
  "atom (x::name) \<sharp> t \<Longrightarrow> atom x \<sharp> (t\<^sup>\<bullet>)"
apply (nominal_induct t avoiding: x rule: term.strong_induct)
apply (auto)
apply (rule_tac y = "term1\<^sup>\<bullet>" and c = "term2\<^sup>\<bullet>" in term.strong_exhaust)
apply (auto simp: fresh_star_def intro: fresh_fact)
done

lemma rhs:
  "subst x (s\<^sup>\<bullet>) (t\<^sup>\<bullet>) \<rightarrow>\<^sub>\<beta>\<^sup>* (subst x s t)\<^sup>\<bullet>"
proof (nominal_induct t avoiding: x s rule: term.strong_induct)
  case (Abs y t)
  moreover then have "atom y \<sharp> (s\<^sup>\<bullet>)" by (auto dest: fresh_atom_bullet)
  ultimately show ?case by (auto dest: fresh_atom_bullet intro: Abs_betas)
next
  case (App t u)
  then show ?case
    unfolding subst.simps
    apply (cases "t\<^sup>\<bullet>" rule: term.strong_exhaust [of _ _ "u\<^sup>\<bullet>"];
           cases "(subst x s t)\<^sup>\<bullet>" rule: term.strong_exhaust [of _ _ "(subst x s u)\<^sup>\<bullet>"])
    apply (auto)
    apply (rule App_betas, simp, metis)
    apply (rule App_betas, simp, metis)
    apply (auto intro: Abs_betas App_betas rtrancl_into_rtrancl split: if_splits)
  show ?case sorry
qed simp

locale z_property =
  fixes bullet :: "'a \<Rightarrow> 'a" ("_\<^sup>\<bullet>")
    and R :: "'a rel"
  assumes Z: "(a, b) \<in> R \<Longrightarrow> (b, a\<^sup>\<bullet>) \<in> R\<^sup>* \<and> (a\<^sup>\<bullet>, b\<^sup>\<bullet>) \<in> R\<^sup>*"
begin

lemma CR: "CR R" sorry

end

end
