(*
Author:  Alexander Lochmann <alexander.lochmann@uibk.ac.at> (2018)
License: LGPL (see file COPYING.LESSER)
*)
theory Ackbo_more
  imports Ackbo_wf
begin

context admissible_weight_fun_ac
begin

abbreviation NS where "NS \<equiv> ackbo_S \<union> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"

lemma ackbo_ce_compat: "ce_trs cn \<subseteq> ackbo_S \<inter> NS"
proof -
  obtain c n where cn: "cn = (c,n)" by force
  {
    fix s t :: "('f,'v) term"
    assume "(s,t) \<in> ce_trs cn"
    hence "s \<rhd> t" unfolding cn ce_trs.simps by auto
    then have "(s, t) \<in> ackbo_S" using ackbo_subterm_property by blast 
    then have "(s,t) \<in> ackbo_S \<inter> NS" by auto
  }
  thus ?thesis by auto
qed

lemma S_ce_compat: "ce_trs cn \<subseteq> ackbo_S" using ackbo_ce_compat by blast
lemma NS_ce_compat: "ce_trs cn \<subseteq> NS" using ackbo_ce_compat by blast

lemma ackbo_redpair: "mono_ce_af_redtriple_order ackbo_S NS NS full_af"
proof (unfold_locales)
  show "SN ackbo_S" using ackbo_strongly_normalizing unfolding SN_defs by blast
  show "subst.closed ackbo_S" using ackbo_subst by blast 
  show "subst.closed NS" by (simp add: \<open>subst.closed ackbo_S\<close> subst.closed_Un subst.closed_conversion subst_closed_rstep) 
  show "ctxt.closed NS" by (metis UnCI UnE ackbo_ctxt case_prodD case_prodI conversion_ctxt_closed ctxt.closedI mem_Collect_eq) 
  show "NS O ackbo_S \<subseteq> ackbo_S" by (simp add: ackbo_ord_p.S_O_S(1) ackbo_ord_p.compat_NS_S) 
  show "ackbo_S O NS \<subseteq> ackbo_S" using \<open>(ackbo_S \<union> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) O ackbo_S \<subseteq> ackbo_S\<close> ackbo_ord_p.S_O_NS(1) by auto  
  show "trans ackbo_S" using S_trans unfolding trans_def by auto
  show "trans NS" by (smt UnCI UnE ackbo_ord_p.compat_NS_S_point ackbo_ord_p.compat_S_NS_point ackbo_ord_p.trans_NS_point ackbo_ord_p.trans_S_point trans_def) 
  show "refl NS" by (simp add: refl_on_def) 
  show "af_compatible full_af NS" by (rule full_af)
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c. ce_trs (c,m) \<subseteq> NS"
    using NS_ce_compat by blast
  show "\<exists> n. \<forall> m \<ge> n. \<forall> c . ce_trs (c,m) \<subseteq> ackbo_S"
    using S_ce_compat by blast
  show "ackbo_S \<subseteq> NS" by auto
qed

lemma ackbo_S_mono: "ctxt.closed ackbo_S"
  using ackbo_ctxt by blast

lemma NS_supteq: "s \<unrhd> t \<Longrightarrow> (s,t) \<in> NS"
  by (metis UnI1 UnI2 ackbo_ord_p.refl_NS_point ackbo_subterm_property case_prodI mem_Collect_eq subterm.dual_order.not_eq_order_implies_strict) 

end
end
