(*
Author:  Alexander Lochmann <alexander.lochmann@uibk.ac.at> (2018)
License: LGPL (see file COPYING.LESSER)
*)
theory Ackbo_split
  imports Ackbo
begin

context admissible_weight_fun_ac
begin

lemma term_size_dec_subterm:
  assumes "s \<rhd> x"
  shows "term_size s > term_size x"
  by (simp add: assms supt_term_size)

lemma term_size_dec_arg:
  assumes "i < length ss"
    and "i < length ts"
  shows "max (term_size (ss ! i)) (term_size (ts ! i)) < max (term_size (Fun f ss)) (term_size (Fun g ts))"
  by (smt assms(1) assms(2) elem_le_sum_list le_imp_less_Suc length_map max.strict_coboundedI1 max.strict_coboundedI2 max_def nth_map term_size.simps(2))

lemma actop_dec_term_size_single:
  assumes "x \<in># actop f (Fun f [s1,s2])"
  shows "term_size x < term_size (Fun f [s1,s2])"
  using assms actop_mset_elem_subterm term_size_dec_subterm
  by (smt Suc_eq_plus1 le_SucE le_add_diff_inverse2 max.coboundedI2 max.commute max_def nat_less_le term_size_at_least1) 

lemma actop_dec_term_size:
  assumes "x \<in># actop f (Fun f [s1,s2])"
    and "y \<in># actop f (Fun f [t1,t2])"
  shows "max (term_size x) (term_size y) \<le> max (term_size (Fun f [s1,s2])) (term_size (Fun f [t1,t2])) - 1"
  using assms actop_mset_elem_subterm term_size_dec_subterm
  by (smt Suc_eq_plus1 le_SucE le_add_diff_inverse2 max.coboundedI2 max.commute max_def nat_less_le term_size_at_least1) 


inductive ackbo_base :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
    weight: "vars_term_ms t \<subseteq># vars_term_ms s \<Longrightarrow> weight s > weight t \<Longrightarrow> ackbo_base s t" |
    base_0: "vars_term_ms t \<subseteq># vars_term_ms s \<Longrightarrow> weight s = weight t \<Longrightarrow>
              ackbo_case0 s t \<Longrightarrow> ackbo_base s t" |
    base_1: "vars_term_ms t \<subseteq># vars_term_ms s \<Longrightarrow> weight s = weight t \<Longrightarrow>
              ackbo_case1 s t \<Longrightarrow> ackbo_base s t"

inductive ackbo_ext :: "nat \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  where
    base_2: "weight s \<le> k \<Longrightarrow> weight t \<le> k \<Longrightarrow> vars_term_ms t \<subseteq># vars_term_ms s \<Longrightarrow> weight s = weight t \<Longrightarrow>
             ackbo_case2 (\<lambda> x y. ackbo_base x y \<or> ackbo_ext (k - 1) x y) s t \<Longrightarrow> ackbo_ext k s t" |
    base_3: "weight s \<le> k \<Longrightarrow> weight t \<le> k \<Longrightarrow> vars_term_ms t \<subseteq># vars_term_ms s \<Longrightarrow> weight s = weight t \<Longrightarrow>
             ackbo_case3 (\<lambda> x y. ackbo_base x y \<or> ackbo_ext (k - 1) x y) s t \<Longrightarrow> ackbo_ext k s t"

lemma ackbo_base_lhs_fun:
  assumes "ackbo_base s t"
  shows "is_Fun s"
  using assms
proof (cases rule: ackbo_base.cases)
  case weight
  then show ?thesis using leD weight_w0 by fastforce
next
  case base_0
  then show ?thesis using ackbo_case0.cases[of s t] by blast
qed auto

lemma ackbo_base_vars_trans[trans]:
  assumes "ackbo_base s t"
    and "ackbo_base t u"
  shows "vars_term_ms u \<subseteq># vars_term_ms s"
proof -
  have "vars_term_ms u \<subseteq># vars_term_ms t" using assms(2) ackbo_base.simps[of t u] by blast
  moreover have "vars_term_ms t \<subseteq># vars_term_ms s" using assms(1) ackbo_base.simps[of s t] by blast
  ultimately show ?thesis using subset_mset.order_trans by blast
qed

lemma ackbo_base_trans[trans]:
  assumes "ackbo_base s t"
    and "ackbo_base t u"
  shows "ackbo_base s u"
  using assms
proof (induct t arbitrary: s u rule:term.induct)
  case (Var x)
  then show ?case using ackbo_base_lhs_fun by blast 
next
  case (Fun g ss)
  note assms = Fun(2,3)
  note IH = Fun(1)
  consider (a) "weight s > weight (Fun g ss) \<and> weight (Fun g ss) > weight u \<or> weight s = weight (Fun g ss) \<and> weight (Fun g ss) > weight u \<or>
                weight s > weight (Fun g ss) \<and> weight (Fun g ss) = weight u" | (b) "weight s = weight (Fun g ss) \<and> weight (Fun g ss) = weight u"
    using ackbo_base.simps assms by blast
  then show ?case
  proof cases
    case a
    then have "weight s > weight u" using less_trans by auto
    then show ?thesis using assms ackbo_base.weight ackbo_base_vars_trans[OF assms] by auto
  next
    case b
    then have f:"is_Fun s" using assms using ackbo_base_lhs_fun by blast 
    then obtain f and ts
      where w:"s = Fun f ts" using f by (meson is_FunE)
    then have "\<not> ackbo_case0 s (Fun g ss)" using ackbo_case0_rhs_var f by auto
    then have "ackbo_case1 s (Fun g ss)"  by (metis Fun.prems(1) ackbo_base.simps b less_not_refl)
    then have st:"pr_strict (f, length ts) (g, length ss)" using ackbo_case1.simps(1) w by simp
    then have "is_Fun u"
    proof -
      consider (c) "w(g, length ss) = 0" |
        (d) "w(g, length ss) \<noteq> 0" by auto
      then show ?thesis
      proof cases
        case c
        then have "1 < length ss" using adm st by (metis leD less_one linorder_neqE_nat pr_str w0(1) w0(2)) 
        then have "\<not> ackbo_case0 (Fun g ss) u" using w by (metis One_nat_def Suc_eq_plus1 ackbo_case0.cases list.size(3) list.size(4) nat_neq_iff term.sel(4))  
        then show ?thesis using b ackbo_base.cases[of "Fun g ss" u] assms(2) by fastforce 
      next
        case d
        have "\<not> ackbo_case0 (Fun g ss) u"
        proof
          assume "ackbo_case0 (Fun g ss) u"
          then obtain x where "u = Var x" using ackbo_case0_rhs_var by blast 
          then have "w(g, length ss) = 0" using w d  weight.simps
            by (metis \<open>ackbo_case0 (Fun g ss) u\<close> ackbo_case0_lhs_length add_le_same_cancel2 b le_zero_eq mult.right_neutral weight_lower_bound_list) 
          then show False using d by simp
        qed
        then show ?thesis using b ackbo_base.cases[of "Fun g ss" u] assms(2) by fastforce 
      qed
    qed
    then obtain h us where w2:"u = Fun h us" using local.ackbo_case1.elims(2)[of "Fun g ss" u] by blast
    then have "pr_strict (g, length ss) (h, length us) \<or> (g, length ss) = (h, length us)"
      using ackbo_base.simps[of "Fun g ss" u] assms(2) b ackbo_case1.simps(1)
      by (metis (no_types, hide_lams) \<open>is_Fun u\<close> ackbo_case0_rhs_var less_not_refl) 
    then have "pr_strict (f, length ts) (h, length us)" using pr_trans st by (metis (no_types, hide_lams)) 
    then show ?thesis using ackbo_base.base_1 b ackbo_case1.simps(1) w2 w ackbo_base_vars_trans[OF assms(1,2)] by fastforce
  qed
qed

lemma ackbo_base_weight_ineq:
  assumes "ackbo_base s t"
    and "weight s \<noteq> weight t"
  shows "weight s > weight t"
  using assms ackbo_base.simps by blast
  
lemma ackbo_01_weight_eq:
  assumes "ackbo_base s t"
    and "\<not> weight s > weight t"
  shows "weight s = weight t"
  using assms
  by (induct rule:ackbo_base.induct) auto

lemma ackbo_01_weight_pos:
  assumes "ackbo_base s t"
  shows "weight s = weight t \<or> weight s > weight t"
  using assms
  by (induct rule:ackbo_base.induct) auto

lemma ackbo_base_irr:
  shows "\<not> ackbo_base s s"
proof
  assume A:"ackbo_base s s"
  then have "\<not> ackbo_case0 s s" using ackbo_base_lhs_fun ackbo_case0_rhs_var by blast
  moreover have "\<not> ackbo_case1 s s" using ackbo_case1.elims(2) pr_irr by blast
  ultimately show "False" using A using ackbo_base.simps by blast
qed

lemma ackbo_base_rel_comp_1:
  assumes "(x,y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
    and "(y,z) \<in> {(x, y). ackbo_base x y}"
  shows "(x,z) \<in> {(x, y). ackbo_base x y}"
  using assms
proof -
  have terms_xy_share_vars:"vars_term_ms x = vars_term_ms y" using ac_terms_share_vars assms(1) by blast
  have "vars_term_ms z \<subseteq># vars_term_ms y" using assms(2) using ackbo_base.simps by blast 
  then have terms_xz_share_vars:"vars_term_ms z \<subseteq># vars_term_ms x" using terms_xy_share_vars by simp 
  have weight_ac_terms:"weight x = weight y" using ac_terms_weight_eq assms(1) by blast
  have "weight y = weight z \<or> weight y > weight z" using ackbo_01_weight_pos assms(2) by blast
  thus ?thesis
  proof 
    assume weight_assms:"weight y = weight z"
    then show "(x, z) \<in> {(x, y). ackbo_base x y}"
    proof (cases "is_Var z")
      case True
      then obtain w where term_z:"z = (Var w)" by blast
      then have not_case2:"\<not> ackbo_case1 y (Var w)" by simp
      then have "ackbo_case0 y (Var w)" using weight_assms ackbo_base.cases[of y "Var w"]
        using assms(2) term_z by fastforce 
      then show ?thesis using assms(1) weight_assms terms_xy_share_vars terms_xz_share_vars
      proof (induct y arbitrary: x)
        case (Var y)
        then show ?case
          by (metis acstep_on_var_impl_var_eq base_0 case_prodI conversion_inv mem_Collect_eq term_z) 
      next
        case (Fun t ts)
        then have r:"root x = Some (t, length ts)" using ac_terms_eq_root[of x "Fun t ts"] by auto
        then have "\<exists> ss. x = Fun t ss" using root_Some by (metis fst_conv)
        then obtain ss where term_1:"x = Fun t ss" by auto
        then have a:"length ss = length ts" using r by simp
        then have "length ss = 1" using Fun.prems(1) ackbo_case0_lhs_length by auto
        then obtain subterm_x where a1:"ss = [subterm_x]" by (metis Suc_eq_plus1 add_cancel_right_left length_0_conv list.size(4) list_splits.cases)
        from a obtain subterm_y where a2:"ts = [subterm_y]" by (metis List.append_eq_append_conv One_nat_def \<open>length ss = 1\<close> add_cancel_right_left append_butlast_last_id list.size(3) list.size(4)) 
        have ac:"(subterm_x, subterm_y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*" using a1 a2 Fun(3) term_1 by (meson acstep_arg_of_unary_fun)
        show ?case
        proof (cases "is_Var subterm_y")
          case True
          then have "is_Var subterm_x" using ac by (metis ac_terms_eq_root is_Fun_Fun_conv root.simps(2) root_Some)
          then have "subterm_x = subterm_y" using ac True ac_terms_share_vars[of subterm_x subterm_y] using ac_vars_eq by auto
          then have "ackbo_case0 x z" using Fun(2) a1 a2 term_1 term_z by blast
          thus ?thesis using Fun(4-6) term_1 by (metis Fun.prems(2) ac_terms_weight_eq base_0 case_prodI mem_Collect_eq) 
        next
          case False
          then obtain us where w:"subterm_y = Fun t us" using Fun(2) ackbo_case0.cases a2 by blast
          then have "weight subterm_y = weight (Fun t ts)" using Fun(4) term_z weight_w0 by (metis a2 add.right_neutral le_add2 le_less list.map(1) list.map(2) not_le sum_list_simps(1) sum_list_simps(2) weight_fun.weight.simps(1) weight_fun.weight.simps(2))  
          moreover have "vars_term_ms subterm_y = vars_term_ms (Fun t ts)" using a2 vars_term_ms.simps(2)[of t ts] by simp 
          moreover have "root x = root subterm_x" using a1 term_1 w ac by (smt False Fun.prems(1) True a a2 ac_terms_eq_root ackbo_case0.cases ackbo_case0_lhs_length root.simps(2) term.sel(4) term_z the_elem_set) 
          moreover have "ackbo_case0 subterm_y (Var w)" using False Fun(2) a2 using ackbo_case0.cases by auto
          ultimately have "ackbo_case0 x z" using a1 term_1 term_z Fun(2-6) Fun(1)[of subterm_y subterm_x]
              ac_terms_share_vars[of subterm_x subterm_y AC]
            by (metis True a2 ac ac_terms_weight_eq ackbo_case0.simps ackbo_rhs_var_imp_case0 admissible_weight_fun_ac.ackbo_base.cases admissible_weight_fun_ac_axioms case_1 list.set_intros(1) mem_Collect_eq nat_less_le prod.simps(2) root.simps(2) root_Some term.inject(2)) 
          then show ?thesis using term_1 Fun(4) Fun(6) Fun(3) ac_share_weight[of x "Fun t ts"]
            by (metis ac_terms_weight_eq base_0 case_prodI mem_Collect_eq) 
        qed
      qed
    next
      case False
      then have "\<not> ackbo_case0 y z" using ackbo_case0_rhs_var by blast
      then have "ackbo_case1 y z" using weight_assms assms(2) by (simp add: ackbo_base.simps)
      then have "ackbo_case1 x z" using assms ac_terms_eq_root[of x y AC]
        by (smt False ackbo_case1.elims(2) ackbo_case1.elims(3) acstep_on_var_impl_var_eq is_Var_def root.elims root_Some term.inject(2))  
      then show ?thesis using weight_assms ac_share_weight[of x y] terms_xz_share_vars by (simp add: base_1 weight_ac_terms) 
    qed
  next
    assume "weight y > weight z"
    then have "weight x > weight z" using weight_ac_terms by simp
    then show ?thesis using terms_xz_share_vars by (simp add: weight) 
  qed
qed

lemma ackbo_base_rel_comp_2:
  assumes "(x,y) \<in> {(x, y). ackbo_base x y}"
    and "(y,z) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
  shows "(x,z) \<in> {(x, y). ackbo_base x y}"
  using assms
proof (cases "is_Var z")
  case True
  then have "y = z" using assms(2) by (metis ac_terms_eq_root ac_vars_eq is_Fun_Fun_conv root.simps(2) root_Some)
  then show ?thesis using assms(1) by simp
next
  case False
  then have "is_Fun y" using assms(2) False by (metis ac_terms_eq_root is_Fun_Fun_conv root.simps(2) root_Some)
  then obtain t ts where term_y:"y = Fun t ts" by blast
  from False obtain us where term_z:"z = Fun t us" using assms(2) term_y ac_terms_eq_root[of y z AC] by auto 
  have "is_Fun x" using assms(1) using ackbo_base_lhs_fun by blast
  then obtain s ss where term_x:"x = Fun s ss" by blast
  then have "pr_strict (s, length ss) (t, length ts) \<and> weight x = weight y \<or> weight x > weight y" using assms(1) term_x term_y
    using ackbo_base.simps ackbo_case0_rhs_var ackbo_case1.simps(1) by blast  
  then show ?thesis
  proof
    assume a:"pr_strict (s, length ss) (t, length ts) \<and> weight x = weight y"
    then have"ackbo_case1 x z" using term_x term_z term_y
      by (metis AC_class_abstract_aocnf.aocnf_roots ackbo_case1.simps(1) aocnf_abstract_aocnf.acconv_imp_nf_eq assms(2) option.inject root.simps(2))
    then show ?thesis using a ac_terms_weight_eq[of y z]
      by (metis ac_terms_share_vars ackbo_base.simps assms(1) assms(2) case_prodD case_prodI mem_Collect_eq) 
  next
    assume a:"weight y < weight x"
    thus ?thesis using ac_terms_weight_eq[of y z] assms(2)
      by (metis ac_terms_share_vars ackbo_base.cases assms(1) mem_Collect_eq old.prod.case weight) 
  qed
qed

interpretation base: order_pair "{(x,y). ackbo_base x y}" "((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
proof 
  show "refl ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)" by (metis conversion_rtrancl refl_rtrancl)
  show "trans {(x, y). ackbo_base x y}" using ackbo_base_trans trans_def by blast
  show "trans ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)" using Abstract_Rewriting.conversion_trans[of "(acstep AC AC)"] by simp
  show "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* O {(x, y). ackbo_base x y} \<subseteq> {(x, y). ackbo_base x y}" using ackbo_base_rel_comp_1 by auto
  show "{(x, y). ackbo_base x y} O (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* \<subseteq> {(x, y). ackbo_base x y}" using ackbo_base_rel_comp_2 by auto
qed

lemma ackbo_ext_0_empty:
  shows "{(x,y). ackbo_ext 0 x y} = {}"
  by (metis (no_types, lifting) Collect_empty_eq ackbo_ext.cases case_prodE le_zero_eq not_less w0(2) weight_w0)

lemma ackbo_ext_incremental:
  assumes "ackbo_ext k s t" 
  shows "ackbo_ext (k + 1) s t"
  using assms
proof (induct k arbitrary: s t)
  case 0
  then show ?case using ackbo_ext_0_empty by auto
next
  case (Suc k)
  then have "ackbo_case2 (\<lambda> x y. ackbo_base x y \<or> ackbo_ext k x y) s t \<or> ackbo_case3 (\<lambda> x y. ackbo_base x y \<or> ackbo_ext k x y) s t"
    using ackbo_ext.cases[of "Suc k" s t] by (metis diff_Suc_1)
  then have "ackbo_case2 (\<lambda> x y. ackbo_base x y \<or> ackbo_ext (Suc k) x y) s t \<or> ackbo_case3 (\<lambda> x y. ackbo_base x y \<or> ackbo_ext (Suc k) x y) s t"
    using Suc(1) by (smt Suc_eq_plus1 ackbo_case2_mono ackbo_case3_mono)
  moreover have "weight s \<le> Suc (Suc k) \<and> weight t \<le> Suc (Suc k)" using Suc(2) ackbo_ext.simps[of "Suc k" s t] by (metis le_SucI)
  moreover have "weight s = weight t \<and> vars_term_ms t \<subseteq># vars_term_ms s" using Suc(2) ackbo_ext.simps[of "Suc k" s t] by auto
  ultimately show ?case using Suc(2) ackbo_ext.intros[of s "Suc (Suc k)" t] diff_Suc_1[of "Suc k"] by auto
qed

lemma ackbo_ext_incremental_set:
  shows "\<And>k. {(x, y). ackbo_ext k x y} \<subseteq> {(x, y). ackbo_ext (k + 1) x y}" 
  using ackbo_ext_incremental by blast

lemma ackbo_ext_upper:
  assumes "ackbo_ext k x y"
    and "k \<le> n"
  shows "ackbo_ext n x y" 
  using assms
proof (induct "n - k" arbitrary: n)
  case 0
  then show ?case by auto 
next
  case (Suc xa)
  then show ?case
  proof (cases "k = n")
    case True
    then show ?thesis using Suc(3) by auto
  next
    case False
    then have "k < n" using Suc(4) by auto
    then have "ackbo_ext (n - 1) x y" using Suc by auto   
    then show ?thesis using ackbo_ext_incremental by (metis One_nat_def Suc_diff_Suc add_diff_inverse_nat diff_is_0_eq le_add1 le_add_diff_inverse2 nat_less_le zero_less_Suc) 
  qed
qed

abbreviation ackbo_wk
  where
    "ackbo_wk k \<equiv> \<lambda> x y. ackbo_ext k x y \<or> ackbo_base x y"

lemma ackbo_wk_incremental:
  assumes "ackbo_wk k s t" 
  shows "ackbo_wk (k + 1) s t"
  using assms ackbo_ext_incremental by blast  

lemma ackbo_ext_impl_case2_or_case3:
  assumes "ackbo_ext (Suc k) s t"
  shows "ackbo_case2 (ackbo_wk k) s t \<or> ackbo_case3 (ackbo_wk k) s t"
  using ackbo_ext.simps[of "Suc k" s t]
  by (smt ackbo_case2_mono admissible_weight_fun_ac.ackbo_case3_mono admissible_weight_fun_ac_axioms assms diff_Suc_1) 

lemma ackbo_wk_not_base_impl_ackbo_ext[simp]:
  assumes "ackbo_wk k x y"
    and "\<not> ackbo_base x y"
  shows "ackbo_ext k x y"
  using assms by fastforce

lemma ackbo_ext_impl_weight_le_k:
  assumes "ackbo_ext k s t"
  shows "weight s \<le> k \<and> weight t \<le> k"
  using assms ackbo_ext.simps[of k s t] by auto

lemma ackbo_ext_impl_vars_subterm:
  assumes "ackbo_ext k s t"
  shows "vars_term_ms t \<subseteq># vars_term_ms s"
  using assms ackbo_ext.simps[of k s t] by auto

lemma ackbo_ext_impl_weight_eq:
  assumes "ackbo_ext k s t"
  shows "weight s = weight t"
  using assms ackbo_ext.simps[of k s t] by auto

lemma ackbo_wk_impl_ackbo:
  assumes "ackbo_wk k x y"
  shows "ackbo x y"
  using assms
proof(induct k arbitrary: x y)
  case 0
  then show ?case
    by (metis ackbo.intros(1) ackbo.intros(2) ackbo.intros(3) ackbo_base.cases ackbo_ext_0_empty empty_iff mem_Collect_eq old.prod.case)  
next
  case (Suc k)
  let ?f = "ackbo_wk k"
  consider (a) "ackbo_base x y"| (b) "ackbo_ext (Suc k) x y" using Suc.prems by blast 
  then show ?case
  proof (cases)
    case a
    then show ?thesis using Suc.hyps by blast
  next
    case b
    then have "ackbo_case2 ?f x y \<or> ackbo_case3 ?f x y" using ackbo_ext.simps[of "Suc k" x y]
      by (smt Un_iff ackbo_case2_mono ackbo_case3_mono case_prodI diff_Suc_1 mem_Collect_eq)
    then show ?thesis
    proof
      assume a:"ackbo_case2 ?f x y"
      then have "ackbo_case2 ackbo x y" using Suc(1) by (smt ackbo_case2_mono) 
      then show "ackbo x y" using ackbo.intros(4) ackbo_ext.simps b by blast 
    next
      assume a:"ackbo_case3 ?f x y"
      then have "ackbo_case3 ackbo x y" by (smt Suc.hyps ackbo_case3_mono) 
      then show "ackbo x y" using ackbo.intros(5) ackbo_ext.simps b by blast  
    qed
  qed
qed

lemma ackbo_impl_ex_k_ackbo_k:
  assumes "ackbo s t"
    and "term_size s + weight s \<le> k"
  shows "ackbo_wk k s t"
  using assms
proof (induct k arbitrary: s t)
  case 0
  then show ?case by (metis add_is_0 le_0_eq not_one_le_zero term_size_at_least1) 
next
  case (Suc k)
  show ?case using Suc(2)
  proof (cases rule: ackbo.cases)
    case case_w
    then show ?thesis using weight by blast
  next
    case case_0
    then show ?thesis using base_0 by blast  
  next
    case case_1
    then show ?thesis using base_1 by blast 
  next
    case case_2
    then obtain f ss ts where wit:"s = Fun f ss \<and> t = Fun f ts" using ackbo_case2_arg_intro by blast
    then have "\<forall> i < length ss. term_size (ss ! i) < term_size s" using Suc term_size_dec_arg_single[of _ ss f] by blast
    then have "\<forall> i < length ss. term_size (ss ! i) + weight (ss ! i) \<le> k" using Suc(3)
      by (smt add_mono_thms_linordered_field(3) elem_le_sum_list le_add2 length_map less_Suc_eq_le max.strict_coboundedI1 max_def nth_map order_trans weight.simps(2) wit)
    then have "\<And> i. i < length ss \<Longrightarrow> i < length ts \<Longrightarrow> ackbo (ss ! i) (ts ! i) \<Longrightarrow> ackbo_wk k (ss ! i) (ts ! i) " using Suc.hyps
      by (meson add_leD2)
    then have "ackbo_case2 (ackbo_wk k) s t" by (smt case_2(3) ackbo_case2.simps(1) ackbo_wk_impl_ackbo lex_ext_cong wit) 
    then have "ackbo_ext (Suc k) s t" using Suc diff_Suc_1 ackbo_ext.intros(1)[of s "Suc k" t]
      by (smt ackbo_case2_mono add_leD2 local.case_2(1) local.case_2(2))
    then show ?thesis by blast
  next
    case case_3
    then obtain f s1 s2 t1 t2 where terms:"s = Fun f [s1,s2] \<and> t = Fun f [t1, t2]" using ackbo_case3_arg_intro by blast
    let ?S = "actop f (Fun f [s1,s2])"
    let ?T = "actop f (Fun f [t1, t2])"
    have "\<And>x. x \<in># ?S \<Longrightarrow> term_size x + weight x \<le> k" using Suc actop_dec_term_size_single
      by (smt ab_semigroup_add_class.add_ac(1) actop_mset_elem_subterm add_mono_thms_linordered_field(3) le_iff_add less_Suc_eq_le subterm_weight terms) 
    then have "\<And>x y. x \<in># ?S \<Longrightarrow> y \<in># ?T \<Longrightarrow> ackbo x y \<Longrightarrow> ackbo_wk k x y" using Suc by blast
    then have "ackbo_case3 (ackbo_wk k) s t" using ackbo_case3_local_mono by (smt case_3 terms)
    then have "ackbo_ext (Suc k) s t" using Suc diff_Suc_1 ackbo_ext.intros(2)[of s "Suc k" t]
      by (smt ackbo_case3_mono add_leD2 local.case_3(1) local.case_3(2)) 
    then show ?thesis by blast
  qed
qed

lemma ackbo_ext_lhs_fun:
  assumes "ackbo_ext k x y"
  shows "is_Fun x"
  using assms ackbo_ext.cases by fastforce

lemma ackbo_ext_rhs_fun:
  assumes "ackbo_ext k x y"
  shows "is_Fun y"
  using assms ackbo_ext.cases by fastforce 

lemma ackbo_ext_eq_root:
  assumes "ackbo_ext k x y"
  shows "root x = root y"
  using assms ackbo_case2_pres_root ackbo_case3_pres_root
proof (cases rule: ackbo_ext.cases)
  case base_2
  then show ?thesis using ackbo_case2_pres_root by blast 
next
  case base_3
  then show ?thesis using ackbo_case3_pres_root by blast 
qed

definition ackbo_wk_rel
  where 
    "ackbo_wk_rel k = {(x,y). ackbo_wk k x y}"

lemma ackbo_wk_rel_eq:
  shows "(x,y) \<in> ackbo_wk_rel k \<longleftrightarrow> ackbo_wk k x y"
  unfolding ackbo_wk_rel_def by blast

lemma ackbo_impl_ackbo_wk:
  fixes x ::"('f,'v) term"
  assumes "(x,y) \<in> {(x,y). ackbo x y}"
  shows "(x,y) \<in> \<Union> {ackbo_wk_rel k | k. True}"
proof -
  have upper_x:"size (vars_term_ms x) \<le> max (size (vars_term_ms x)) (size (vars_term_ms y))" by simp 
  have upper_y:"size (vars_term_ms y) \<le> max (size (vars_term_ms x)) (size (vars_term_ms y))" by simp
  from assms have "ackbo x y" by simp
  then have "\<exists> k. ackbo_wk k x y" using ackbo_impl_ex_k_ackbo_k le_add1 le_add2 by blast 
  then show ?thesis using ackbo_wk_rel_eq by blast
qed

lemma ackbo_subset_ackbo_wk:
  shows "{(x,y). ackbo x y} \<subseteq> \<Union> {ackbo_wk_rel k | k. True}"
  using ackbo_impl_ackbo_wk by blast

lemma ackbo_wk_subset_ackbo:
  shows "\<Union> {ackbo_wk_rel k | k. True} \<subseteq> {(x,y). ackbo x y}"
  using ackbo_wk_impl_ackbo ackbo_wk_rel_eq by blast

lemma ackbo_eq_ackbo_wk_all_k:
  shows "{(x,y). ackbo x y} = \<Union> {ackbo_wk_rel k | k. True}"
  using ackbo_subset_ackbo_wk ackbo_wk_subset_ackbo by blast

lemma ackbo_eq_ackbo_wk:
  shows "ackbo x y = (\<exists> k. ackbo_wk k x y)"
  using ackbo_eq_ackbo_wk_all_k ackbo_wk_rel_eq by blast

lemma ackbo_01_root_trans:
  assumes "ackbo_base x y"
    and "is_Fun y"
    and "root y = root z"
    and "weight z = weight y"
    and "weight x = weight y"
    and "vars_term_ms z \<subseteq># vars_term_ms y"
  shows "ackbo_base x z"
  using assms
proof -
  from assms(2) obtain t ts where term_y:"y = Fun t ts" by blast
  then obtain us where term_z:"z = Fun t us" using assms(3) by (metis fst_conv root.simps(2) root_Some)
  from assms(1) obtain s ss where term_x:"x = Fun s ss" using ackbo_base_lhs_fun by blast
  from assms(1) have "pr_strict (s, length ss) (t, length ts)" using assms(5) term_x term_y
    by (metis ackbo_base.cases ackbo_case0_rhs_var ackbo_case1.simps(1) assms(2) less_not_refl)
  then have "pr_strict (s, length ss) (t, length us)" using assms(3) term_z using term_y by auto
  then show ?thesis using assms term_x term_z  by (metis ackbo_base.simps ackbo_case1.simps(1) subset_mset.order.trans)
qed

lemma ackbo_base_O_wk_impl_ackbo_base:
  assumes "ackbo_base s t"
    and "ackbo_wk k t u"
  shows "ackbo_base s u"
proof -
  have v:"vars_term_ms u \<subseteq># vars_term_ms t" using assms ackbo_ext.simps ackbo_base.simps
  proof -
    have "ackbo t u"
      by (meson ackbo_wk_impl_ackbo assms(2))
    then show ?thesis
      using ackbo.simps by auto
  qed
  then have vars:"vars_term_ms u \<subseteq># vars_term_ms s" using assms ackbo_base.simps by (metis subset_mset.dual_order.trans)
  from assms(2) have "ackbo_base t u \<or> ackbo_ext k t u" by blast 
  then show ?thesis
  proof
    assume "ackbo_base t u"
    then show ?thesis using assms(1) base.trans_S_point by blast 
  next
    assume a:"ackbo_ext k t u"
    then have "root t = root u" using ackbo_ext_eq_root by blast 
    moreover have "is_Fun t" using a ackbo_ext_lhs_fun by blast
    ultimately show ?thesis using a assms v ackbo_01_root_trans[of s t u] ackbo_ext.cases
      by (metis (no_types, lifting) ackbo_01_weight_pos vars weight)
  qed
qed

lemma ackbo_wk_O_base_impl_ackbo_base:
  assumes "ackbo_wk k s t"
    and "ackbo_base t u"
  shows "ackbo_base s u"
proof -
  have v:"vars_term_ms u \<subseteq># vars_term_ms t" using assms ackbo_base.simps by auto
  have v2:"vars_term_ms t \<subseteq># vars_term_ms s" using assms ackbo_base.simps ackbo_ext.simps by blast 
  then have vars:"vars_term_ms u \<subseteq># vars_term_ms s" using v by simp
  from assms(1) have "ackbo_base s t \<or> ackbo_ext k s t" by blast
  then show ?thesis
  proof
    assume "ackbo_base s t"
    then show ?thesis using assms(2) base.trans_S_point by blast 
  next
    assume ackbo_rule23:"ackbo_ext k s t"
    then have eq_weight:"weight s = weight t" using assms(1) using ackbo_ext.cases by blast
    have eq_root:"root s = root t" using ackbo_rule23 ackbo_ext_eq_root by blast
    have "weight t > weight u \<or> weight t = weight u" using assms(2) using ackbo_01_weight_eq by blast 
    then show ?thesis
    proof
      assume "weight u < weight t"
      then show ?thesis using eq_weight vars by (simp add: weight)
    next
      assume weight_assm:"weight t = weight u"
      then show ?thesis
      proof(cases "is_Var u")
        case True
        then have r:"ackbo_case0 t u" using assms(2) weight_assm v using ackbo_base.cases[of t u]
          by (metis ackbo_case1_rhs_Fun nat_less_le)
        then have "num_args t = 1" by (metis ackbo_case0.cases ackbo_case0_lhs_length is_FunI term.collapse(2)) 
        then have not_ackbo_k_case3:"\<not> ackbo_case3 (ackbo_wk (k-1)) s t" by (simp add: ackbo_case3_num_args_2)
        have ackbo_k_case2:"ackbo_case2 (ackbo_wk (k-1)) s t" using ackbo_rule23 not_ackbo_k_case3
        proof cases
          case base_2
          then show ?thesis by (smt ackbo_case2_mono) 
        next
          case base_3
          then show ?thesis using not_ackbo_k_case3 by (smt ackbo_case3_mono) 
        qed
        from not_ackbo_k_case3 have not_ackbo_case3:"\<not> ackbo_case3 ackbo s t" by (smt ackbo_case2.elims(2) ackbo_case3.simps(1) ackbo_k_case2) 
        have ackbo_case2:"ackbo_case2 ackbo s t" using ackbo_rule23 ackbo_k_case2 by (smt ackbo_case2_mono ackbo_wk_impl_ackbo) 
        then show ?thesis using r eq_weight weight_assm v2 
        proof (induct s arbitrary:t)
          case (Var x)
          then show ?case by auto
        next
          case (Fun f fs)
          then obtain t1 where term_t:"t = Fun f [t1]" by (metis (no_types, hide_lams) ackbo_case2_pres_root ackbo_case_0_lhs_subterm_exists root.simps(2) root_Some term.sel(4)) 
          then obtain f1 where subterm_f:"fs = [f1]" using Fun
            by (metis List.append_eq_append_conv ackbo_case2_pres_root append_butlast_last_id list.size(4) root.simps(2) root_Some term.inject(2))
          then have sub_weight:"weight f1 = w0" using Fun(3-5) True eq_iff weight_w0 by fastforce
          have weight_f_0:"w(f,1) = 0" using Fun(3) Fun(5) term_t
            by (metis True ackbo_case0_lhs_length add_cancel_right_left is_Var_def le_add2 le_antisym mult.right_neutral weight_fun.weight.simps(1) weight_fun.weight.simps(2) weight_lower_bound_list)  
          have sub_vars:"vars_term_ms t1 \<subseteq># vars_term_ms f1" using term_t subterm_f Fun(6) by simp
          from Fun(2) have rec_call:"ackbo f1 t1" using term_t subterm_f ackbo_case2_single_arg[of ackbo f f1 t1] by simp
          show ?case
          proof (cases "is_Var t1")
            case True
            then have "t1 = u" using Fun(3) term_t using adm pr_irr weight_f_0
              by (smt Term.term.simps(4) ackbo_case0.simps list.inject term.collapse(1) term.sel(4))
            then have c:"ackbo_case0 f1 u" using rec_call ackbo_rhs_var_imp_case0 sub_weight weight_w0 eq_iff True by fastforce
            have "weight (Fun f [f1]) = w0" using term_t subterm_f weight_f_0 unique_unary_function_weight0[of f] Fun(4) True
              by (metis Fun.prems(4) \<open>t1 = u\<close> is_Var_def weight.simps(1))
            then have "weight f1 = w0" using weight_f_0 by auto
            then have "root f1 = Some (f,1)" using unique_unary_function_weight0[OF weight_f_0] c
              by (metis Fun.prems(3) Fun.prems(4) \<open>t1 = u\<close> \<open>weight (Fun f [f1]) = w0\<close> ackbo_base_lhs_fun ackbo_case0_lhs_length add_le_same_cancel2 base_0 is_FunE le_0_eq mult.comm_neutral root.simps(2) sub_vars subterm_f weight.simps(2) weight_lower_bound_list) 
            then have "ackbo_case0 (Fun f fs) u" using c subterm_f by (metis ackbo_case0.simps fst_conv root_Some)
            then show ?thesis using Fun(2-6) using ackbo_case0_lhs_contains_var base_0 by fastforce 
          next
            case False
            have e:"vars_term_ms u \<noteq> {#}" using True by force
            have p:"vars_term_ms u \<subseteq># vars_term_ms t" using Fun(3) ackbo_case0_lhs_contains_var by blast
            then have "vars_term_ms t \<subseteq># vars_term_ms f1" using term_t subterm_f Fun(6) by simp 
            then have not_empty:"vars_term_ms f1 \<noteq> {#}" using e p by fastforce
            obtain t11 where sub_sub_term:"t1 = Fun f [t11]" using term_t Fun(3) False 
              by (smt True ackbo_case0.simps list.inject term.inject(2))
            moreover have "\<not> ackbo_case0 f1 t1" using sub_sub_term ackbo_case0_rhs_var by blast
            moreover have "\<not> ackbo_case1 f1 t1"
            proof(rule ccontr)
              assume "\<not> \<not> ackbo_case1 f1 t1"
              then have a:"ackbo_case1 f1 t1" by simp
              then obtain l ls where w:"f1 = Fun l ls" using ackbo_case1.elims(2) by blast
              then have "length ls > 0" using not_empty vars_term_not_empty_imp_no_constant by simp   
              then have "length ls = 1" using sub_weight w
                by (metis comm_monoid_mult_class.mult_1 leD less_not_refl2 less_one linorder_neqE_nat mult_right_le_imp_le w0(2) weight_lower_bound_n_arry_fun) 
              then have not_eq_sub_root:"l \<noteq> f" using sub_sub_term a w by (metis Fun.prems(2) ackbo_case0_lhs_length ackbo_case1_different_Fun list.size(4) term_t)  
              then have "w (l, 1) = 0" using sub_weight
                by (metis \<open>length ls = 1\<close> add_cancel_right_left le_add2 le_antisym mult.right_neutral w weight.simps(2) weight_lower_bound_list)
              then show False using not_eq_sub_root weight_f_0 adm by (metis fst_conv pr_str) 
            qed
            moreover have "\<not> ackbo_case3 ackbo f1 t1" using sub_sub_term ackbo_case3.elims(2) by fastforce
            moreover have "weight f1 = weight t1" using term_t subterm_f Fun(4) True by simp 
            ultimately have w:"ackbo_case2 ackbo f1 t1" using rec_call ackbo.simps[of f1 t1] by auto
            then have eq_root_subterms:"root f1 = root t1" using ackbo_case2_pres_root by blast 
            have "ackbo_case0 t1 u" using sub_sub_term Fun(3) term_t ackbo_case0.cases by blast
            moreover have "weight t1 = w0" using \<open>weight f1 = weight t1\<close> sub_weight by linarith 
            ultimately have "ackbo_base f1 u" using Fun(1)[of f1 t1] subterm_f w sub_vars sub_weight True by auto
            then have "ackbo_case0 f1 u" using True \<open>weight f1 = w0\<close> using ackbo_base.cases by fastforce 
            then show ?thesis using subterm_f Fun(4-6) p eq_root_subterms term_t True
              by (metis (no_types, lifting) ackbo_case0.intros(2) ackbo_case2_arg_intro base_0 sub_sub_term subset_mset.dual_order.trans term.inject(2) w)   
            qed
          qed
      next
        case False
        then have "ackbo_case1 t u" using assms(2) weight_assm by (metis ackbo_base.simps ackbo_case0_rhs_var less_not_refl)
        from ackbo_case1.elims(2)[OF this, of "ackbo_case1 s u"]
        have "ackbo_case1 s u" using root.elims[OF eq_root]
          by (metis ackbo_case1.simps(1) option.distinct(1) option.inject root.simps(2))
        moreover have "vars_term_ms u \<subseteq># vars_term_ms s" using v v2 by fastforce  
        ultimately show ?thesis using weight_assm eq_weight by (simp add: base_1) 
      qed
    qed
  qed
qed

lemma comp_base_wk:
  shows "({(x, y). ackbo_base x y} O (ackbo_wk_rel k)) \<union> ((ackbo_wk_rel k) O {(x, y). ackbo_base x y}) \<subseteq> {(x, y). ackbo_base x y}"
proof -
  have "\<And> x y.(x,y) \<in> ackbo_wk_rel k \<longleftrightarrow> ackbo_wk k x y" unfolding ackbo_wk_rel_def by blast
  moreover have "\<And> x y.(x,y) \<in> {(s,t). ackbo_base s t}  \<longleftrightarrow> ackbo_base x y" by blast
  ultimately show ?thesis using ackbo_wk_O_base_impl_ackbo_base ackbo_base_O_wk_impl_ackbo_base
    by (smt Un_iff relcompE subsetI) 
qed

lemma ackbo_wk_trans[trans]:
  fixes s :: "('f, 'v) term"
  assumes "order_pair (ackbo_wk_rel k :: ('f, 'v) Term.term rel) ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "ackbo_wk (Suc k) s t"
    and "ackbo_wk (Suc k) t u"
  shows "ackbo_wk (Suc k) s u"
proof -
  let ?s_rel = "ackbo_wk_rel k :: ('f, 'v) term rel"
  let ?str ="ackbo_wk k :: ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  let ?ns_rel ="(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f, 'v) term rel"
  consider (a) "ackbo_base s t" | (b) "ackbo_base t u"|
    (c) "\<not> ackbo_base s t \<and> \<not> ackbo_base t u" using assms(2-3) by blast
  then show ?thesis
  proof cases
    case a
    then show ?thesis using ackbo_base_O_wk_impl_ackbo_base ackbo_base_trans assms(3) by blast
  next
    case b
    then show ?thesis using ackbo_wk_O_base_impl_ackbo_base ackbo_base_trans assms(2) by blast
  next
    case c
    consider (i) "ackbo_case2 ?str s t" | (ii) "ackbo_case3 ?str s t"
      using c assms(2) ackbo_ext.cases using ackbo_ext_impl_case2_or_case3 by blast 
    then show ?thesis
    proof cases
      case i
      then obtain f ss ts where term_st:"s = Fun f ss \<and> t = Fun f ts" using ackbo_case2_arg_intro by blast
      then have "f \<notin> AC \<or> length ts \<noteq> 2" using ackbo_case2_impl_not_acsym i by auto
      then have "\<not> ackbo_case3 ?str t u" using term_st ackbo_case3.elims(2)[of ?str t u] by fastforce
      then have l:"ackbo_case2 ?str t u" using ackbo_ext_impl_case2_or_case3[of k t u] assms(3) c by auto
      then have c2:"ackbo_case2 (ackbo_wk k) s u" using i l order_pair_impl_ackbo_case2_trans[of "ackbo_wk k" s t u] assms(1) by (metis ackbo_wk_rel_def) 
      have w:"weight s \<le> Suc k" using assms(2) c ackbo_ext.simps by blast
      have eqw:"weight s = weight u" using assms(2,3) c ackbo_ext_impl_weight_eq by metis
      from w eqw have w2:"weight u \<le> Suc k" by simp
      have "vars_term_ms u \<subseteq># vars_term_ms s" using assms(2,3) c ackbo_ext_impl_vars_subterm subset_mset.order.trans by blast 
      from ackbo_ext.intros(1)[OF w w2 this eqw] have "ackbo_ext (Suc k) s u" using c2
        by (smt ackbo_case2_mono diff_Suc_1)
      then show ?thesis by blast
    next
      case ii
      obtain f s1 s2 t1 t2 where terms_st:"s = Fun f [s1,s2] \<and> t = Fun f [t1,t2]" using ii ackbo_case3_arg_intro by blast
      have "f \<in> AC" using ii terms_st by (smt ackbo_case3.simps(1))
      then have "\<not> ackbo_case2 ?str t u" using terms_st ackbo_case2.elims(2) by force
      then have cas:"ackbo_case3 ?str t u" using ii c assms(2-3) ackbo_ext_impl_case2_or_case3 by auto
      then have c3:"ackbo_case3 ?str s u" using order_pair_impl_ackbo_case3_trans[of "ackbo_wk k" s t u] assms(1)
        by (simp add: ackbo_wk_rel_def ii)
            have w:"weight s \<le> Suc k" using assms(2) c ackbo_ext.simps by blast
      have eqw:"weight s = weight u" using assms(2,3) c ackbo_ext_impl_weight_eq by metis
      from w eqw have w2:"weight u \<le> Suc k" by simp
      have "vars_term_ms u \<subseteq># vars_term_ms s" using assms(2,3) c ackbo_ext_impl_vars_subterm subset_mset.order.trans by blast
      from ackbo_ext.intros(2)[OF w w2 this eqw] have "ackbo_ext (Suc k) s u" using c3 by (smt ackbo_case3_mono diff_Suc_1)
      then show ?thesis by blast
    qed
  qed
qed

lemma ackbo_wk_compat_l:
  fixes s :: "('f, 'v) term"
  assumes "order_pair (ackbo_wk_rel k :: ('f, 'v) Term.term rel) ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f, 'v) Term.term rel)"
    and "(s, t) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "ackbo_wk (Suc k) t u"
  shows "ackbo_wk (Suc k) s u"
proof -
  let ?s_rel = "ackbo_wk_rel k :: ('f, 'v) term rel"
  let ?str ="ackbo_wk k :: ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  let ?ns_rel ="(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f, 'v) term rel"
  have ord:"order_pair {(x,y). ?str x y} ?ns_rel" using assms(1) by (simp add: ackbo_wk_rel_def) 
  consider (a) "ackbo_base t u" |
    (b) "\<not> ackbo_base t u" using assms(3) by blast
  then show ?thesis
  proof cases
    case a
    then have "ackbo_base s u" using assms(2) base.compat_NS_S_point by blast 
    then show ?thesis using  assms(3) by blast
  next
    case b
    consider (i) "ackbo_case2 ?str t u" | (ii) "ackbo_case3 ?str t u"
      using b assms(3) ackbo_ext.cases ackbo_ext_impl_case2_or_case3 by blast 
    then show ?thesis
    proof cases
      case i
      then have w:"ackbo_case2 ?str s u" using order_pair_impl_ackbo_case2_l_comp[of ?str s t u] ord assms(2) i by (metis ackbo_wk_rel_def)
      have "term_size s = term_size t" using ac_terms_size_eq assms(2) by auto
      then have "weight s \<le> Suc k" using assms(3) b
        by (metis ac_terms_weight_eq ackbo_ext_impl_weight_le_k assms(2)) 
      moreover have "weight s = weight u" using ac_terms_weight_eq[of s t] ackbo_ext_impl_weight_eq[of "Suc k" t u] b assms(2-3) by auto
      moreover have "vars_term_ms u \<subseteq># vars_term_ms s" using ac_terms_share_vars[of s t AC] ackbo_ext_impl_vars_subterm[of "Suc k" t u] b assms(2-3) by auto 
      ultimately have "ackbo_ext (Suc k) s u" using w base_2[of s "Suc k" u]
        by (smt ackbo_case2_mono diff_Suc_1) 
      then show ?thesis by blast
    next
      case ii
      then have w:"ackbo_case3 ?str s u" using order_pair_impl_ackbo_case3_l_comp[of "ackbo_wk k" s t u] ord assms(2)
        by (simp add: ackbo_wk_rel_def ii)
      have "term_size s = term_size t" using ac_terms_size_eq assms(2) by auto
      then have "weight s \<le> Suc k" using assms(3) b
        by (metis ac_terms_weight_eq ackbo_ext_impl_weight_le_k assms(2)) 
      moreover have "weight s = weight u" using ac_terms_weight_eq[of s t] ackbo_ext_impl_weight_eq[of "Suc k" t u] b assms(2-3) by auto
      moreover have "vars_term_ms u \<subseteq># vars_term_ms s" using ac_terms_share_vars[of s t AC] ackbo_ext_impl_vars_subterm[of "Suc k" t u] b assms(2-3) by auto 
      ultimately have "ackbo_ext (Suc k) s u" using w base_3[of s "Suc k" u] by (smt ackbo_case3_mono diff_Suc_1) 
      then show ?thesis by blast
    qed
  qed
qed

lemma ackbo_wk_compat_r:
  fixes s :: "('f, 'v) term"
  assumes "order_pair (ackbo_wk_rel k :: ('f, 'v) Term.term rel) ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    and "ackbo_wk (Suc k) s t"
    and "(t, u) \<in> ((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  shows "ackbo_wk (Suc k) s u"
proof -
  let ?s_rel = "ackbo_wk_rel k :: ('f, 'v) term rel"
  let ?str ="ackbo_wk k :: ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
  let ?ns_rel ="(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f, 'v) term rel"
  have ord:"order_pair {(x,y). ?str x y} ?ns_rel" using assms(1) by (simp add: ackbo_wk_rel_def) 
  consider (a) "ackbo_base s t" |
    (b) "\<not> ackbo_base s t" using assms(2) by blast
  then show ?thesis
  proof cases
    case a
    then have "ackbo_base s u" using assms(3) base.compat_S_NS_point by blast 
    then show ?thesis using assms(2) by blast
  next
    case b
    consider (i) "ackbo_case2 ?str s t" | (ii) "ackbo_case3 ?str s t"
      using b assms(2) ackbo_ext.cases ackbo_ext_impl_case2_or_case3 by blast 
    then show ?thesis
    proof cases
      case i
      have w:"ackbo_case2 ?str s u" using i order_pair_impl_ackbo_case2_r_comp[of ?str s t u] ord assms(3) by (metis ackbo_wk_rel_def)
      have "term_size u = term_size t" using assms(3) ac_terms_size_eq[of t u] by auto
      then have "weight s \<le> Suc k" using b ackbo_ext_impl_weight_le_k assms(2) by blast 
      moreover have "weight s = weight u" using ac_terms_weight_eq[of t u] ackbo_ext_impl_weight_eq[of "Suc k" s t] b assms(2-3) by auto
      moreover have "vars_term_ms u \<subseteq># vars_term_ms s" using ac_terms_share_vars[of t u AC] ackbo_ext_impl_vars_subterm[of "Suc k" s t] b assms(2-3) by auto 
      ultimately have "ackbo_ext (Suc k) s u" using w base_2[of s "Suc k" u] by (smt ackbo_case2_mono diff_Suc_1)
      then show ?thesis by blast
    next
      case ii
      then have w:"ackbo_case3 ?str s u" using ii order_pair_impl_ackbo_case3_r_comp[of "ackbo_wk k" s t u] ord assms(3) by blast
      have "term_size t = term_size u" using ac_terms_size_eq assms(3) by auto
      then have "weight s \<le> Suc k" using assms(2) b ackbo_ext_impl_weight_le_k by blast 
      moreover have "weight s = weight u" using ac_terms_weight_eq[of t u] ackbo_ext_impl_weight_eq[of "Suc k" s t] b assms(2-3) by auto
      moreover have "vars_term_ms u \<subseteq># vars_term_ms s" using ac_terms_share_vars[of t u AC] ackbo_ext_impl_vars_subterm[of "Suc k" s t] b assms(2-3) by auto 
      ultimately have "ackbo_ext (Suc k) s u" using w base_3[of s "Suc k" u] diff_Suc_1[of k] by (smt ackbo_case3_mono diff_Suc_1)
      then show ?thesis by blast
    qed
  qed
qed

interpretation ackbo_wk_ord_p:
  order_pair "ackbo_wk_rel k :: ('f, 'v) term rel" "((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*) :: ('f, 'v) term rel" for k
proof (induct k)
  case 0
  then have "ackbo_wk_rel 0 = {(x :: ('f, 'v) term ,y). ackbo_base x y}" using ackbo_ext_0_empty ackbo_wk_rel_def by blast
  then show ?case using base.order_pair_axioms by fastforce
next
  case (Suc k)
  let ?str = "ackbo_wk_rel k ::('f, 'v) term rel"
  let ?str' = "ackbo_wk_rel (Suc k) ::('f, 'v) term rel"
  let ?ns = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>* :: ('f, 'v) term rel"
  have IH:"order_pair ?str ?ns" using Suc by simp
  show "order_pair ?str' ?ns"
  proof
    show "refl ?ns" by (simp add: base.refl_NS) 
    show "trans ?ns" using conversion_trans by blast
    show "trans ?str'" using IH ackbo_wk_trans[of k] trans_def ackbo_wk_rel_def by (smt ackbo_wk_rel_eq) 
    show "?ns O ?str' \<subseteq> ?str'" using IH ackbo_wk_compat_l ackbo_wk_rel_def ackbo_wk_rel_eq by blast 
    show "?str' O ?ns \<subseteq> ?str'" using IH ackbo_wk_compat_r ackbo_wk_rel_def ackbo_wk_rel_eq by blast 
  qed
qed

lemma ackbo_wk_rel_incremental:
  assumes "(x,y) \<in> ackbo_wk_rel k"
  shows "(x,y) \<in> ackbo_wk_rel (Suc k)"
  unfolding ackbo_wk_rel_def using ackbo_wk_incremental ackbo_wk_rel_eq assms
  by (metis (no_types, lifting) Suc_eq_plus1 case_prodI mem_Collect_eq)

interpretation ackbo_wk_union_ord_p: order_pair "\<Union> {ackbo_wk_rel k | k. True}" "((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  using ackbo_wk_rel_incremental ackbo_wk_ord_p.order_pair_axioms order_pair_subset_rel[of ackbo_wk_rel "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"] by fastforce

sublocale ackbo_ord_p: order_pair "{(x,y). ackbo x y}" "((acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
  using ackbo_wk_union_ord_p.order_pair_axioms and ackbo_wk_rel_def
  by (simp add: ackbo_eq_ackbo_wk_all_k)

lemma ackbo_wk_irr:
  shows "irrefl (ackbo_wk_rel k)"
  unfolding irrefl_def
proof (induct k)
  case 0
  then show ?case using ackbo_base_irr ackbo_ext_0_empty ackbo_wk_rel_eq by blast 
next
  case (Suc k)
  then have "\<forall> s. \<not> ackbo_base s s" using ackbo_base_irr by auto
  moreover have "\<And> s :: ('f, 'a)term. \<not> ackbo_ext (Suc k) s s"
  proof
    let ?str = "(\<lambda> x y. ackbo_ext k x y \<or> ackbo_base x y)"
    let ?ord = "\<lambda> x y.(?str x y, (x, y) \<in> (acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*)"
    fix s :: "('f, 'a)term"
    assume ass: "ackbo_ext (Suc k) s s"
    then obtain f ss where term_a: "s = Fun f ss" using ackbo_ext_rhs_fun by blast 
    {assume "fst (lex_ext ?ord (length ss) ss ss)"
      then obtain i where "i < length ss" and "?str (ss ! i) (ss ! i)"
        unfolding lex_ext_iff by auto
      then have "ackbo_wk k (ss ! i) (ss ! i)" using ackbo_base_irr by auto
      then have "(ss ! i, ss ! i) \<in> ackbo_wk_rel k" using ackbo_wk_rel_eq[symmetric] by auto
      then have False using Suc by auto}
    then have "\<not> ackbo_case2 ?str s s" using term_a by auto
    moreover have "\<not> ackbo_case3 ?str s s"
    proof
      assume ass: "ackbo_case3 (ackbo_wk k) s s"
      let ?S = "actop f (Fun f ss)"
      let ?s = "ackbo_wk_rel k :: (('f, 'a) Term.term \<times> ('f, 'a) Term.term) set"
      let ?ns = "(acstep AC AC)\<^sup>\<leftrightarrow>\<^sup>*"
      have "?ns O ?s \<subseteq> ?s" using ackbo_wk_ord_p.NS_O_S(1) ackbo_wk_rel_def by blast
      moreover have "?s O ?ns \<subseteq> ?s" using ackbo_wk_ord_p.S_O_NS(1) ackbo_wk_rel_def by blast 
      moreover have "irrefl ?s" using Suc unfolding irrefl_def by auto
      moreover have "trans ?s" by (simp add: ackbo_wk_ord_p.trans_S) 
      ultimately have "\<not> ackbo_case3_filtered_rel ?S ?S (f,length ss) ?ns ?s" by (metis (no_types, lifting) add_diff_cancel_right' irrefl_def order_less_irrefl s_mul_irrefl) 
      then show False using ass ackbo_case3.simps by (smt ackbo_case3_Elim ackbo_case3_mono ackbo_wk_rel_def term_a) 
    qed
    ultimately show False using ass ackbo_ext_impl_case2_or_case3 by blast  
  qed
  ultimately show ?case using ackbo_wk_rel_eq by blast
qed

lemma ackbo_irr:
  shows "irrefl {(s,t). ackbo s t}"
  using ackbo_wk_irr by (metis ackbo_eq_ackbo_wk ackbo_wk_rel_eq irrefl_def mem_Collect_eq old.prod.case) 

end
end